"""
Hold constant values
"""
import os
import matplotlib.colors as mcolors
from rdkit.Chem import AllChem, RDConfig

feature_factory = AllChem.BuildFeatureFactory(os.path.join(RDConfig.RDDataDir, 'BaseFeatures.fdef'))

FEATURES = {
    "Donor": ["[#7!H0&!$(N-[SX4](=O)(=O)[CX4](F)(F)F)]",  # nitrogen in rings or amines
              "[#8!H0&!$([OH][C,S,P]=O)]",  # oxygen atom bonded to hydrogen
              "[#16!H0]"  # sulfur atom
              ],
    "Acceptor": ["[#7&!$([nX3])&!$([NX3]-*=[!#6])&!$([NX3]-[a])&!$([NX4])&!$(N=C([C,N])N)]",
                 # amines or nitrogen in rings non aromatic rings
                 "[$([O])&!$([OX2](C)C=O)&!$(*(~a)~a)]"  # hydroxyl or ether
                 ],
    "Aromatic": ["a1aaaaa1",  # six membered aromatic ring
                 "a1aaaa1",  # five membered aromatic ring
                 "[#6]1[#6]=[#6][#6]=[#7]1",  # fused pyrole ring
                 "[#6]:1:[#6]:[#6]:[#6]:[#6]:[#7]:1"  # fused pyridine ring
                 ],
    "Hydrophobe": ["a1aaaaa1",  # six member aromatic ring
                   "a1aaaa1",  # five member aromatic ring
                   "*~1~*~*~*~*~*~1",  # fix member ring, any atom, any bond
                   "*~1~*~*~*~*~1",  # five member ring, any atom, any bond
                   # methyl, methylene, methine, halogens
                   "[$([CH3X4,CH2X3,CH1X2,F,Cl,Br,I])&!$(**[CH3X4,CH2X3,CH1X2,F,Cl,Br,I])]",
                   # matches to methyl, methylene, terminal methine
                   "[$(*([CH3X4,CH2X3,CH1X2,F,Cl,Br,I])[CH3X4,CH2X3,CH1X2,F,Cl,Br,I])&!$(*([CH3X4,CH2X3,CH1X2,F,Cl,Br,I])([CH3X4,CH2X3,CH1X2,F,Cl,Br,I])[CH3X4,CH2X3,CH1X2,F,Cl,Br,I])]([CH3X4,CH2X3,CH1X2,F,Cl,Br,I])[CH3X4,CH2X3,CH1X2,F,Cl,Br,I]",
                   "[CH3]",  # terminal methyl group
                   "[CH2]~*~!@[*1]",  # terminal methylene group not in a ring
                   ]
}

FEATURE_COLORS = {
    "Donor": (0.2549019607843137, 0.4117647058823529, 0.8823529411764706),  # royalblue
    "Acceptor": (1.0, 0.27058823529411763, 0.0),  # orangered
    "Aromatic": (0.8549019607843137, 0.6470588235294118, 0.12549019607843137),  # goldenrod
    "Hydrophobe": (0.1803921568627451, 0.5450980392156862, 0.3411764705882353),  # seagreen
    "LumpedHydrophobe": (0.1803921568627451, 0.5450980392156862, 0.3411764705882353),  # seagreen
    "PosIonizable": (0.0, 0.7490196078431373, 1.0)  # deepskyblue
}


def color_convert(color: str = None):
    """
    helper funciton to convert color to rgb.
    :param color: str
        Color to convert to rgb. Can be hex or color name.
    :return:
    """
    try:
        # convert color to rgb
        rgb = mcolors.to_rgb(color)
        return rgb
    except:
        raise ValueError(f"{color} is not a valid color!")


if __name__ == "__main__":
    import doctest

    doctest.testmod()



import warnings
import collections
import pandas as pd
import numpy as np
from tqdm import tqdm
from rdkit import Chem
from rdkit.Chem import AllChem
from typing import Union, Optional

from pharmacophore.constants import feature_factory, FEATURES, FEATURE_COLORS, color_convert


class Pharmacophore:
    def __init__(self, features: Union[str, dict] = 'default'):
        """
        Initialize the Pharmacophore class.
        :param features: Union[str, dict]
            Set features to use. Will default to 'default'. Can also accept RDKit features or a dictionary of features.
        """
        self.sdf = None
        self.features = features
        self.features_list = None

    def read_sdf(self, sdf_file: str = None, verbose: bool = False):
        """
        Read sdf files.
        :param sdf_file: Str
            File path to .sdf file.
        :param verbose: Bool
            Output description when reading .sdf file.
        :return: list
            A list of molecules in Chem.Mol format.
        """
        supplier = Chem.SDMolSupplier(sdf_file)
        # extract ROMols into a list
        mol_list = []
        if verbose:
            for mol in tqdm(supplier, desc="Reading Molecules"):
                mol_list.append(mol)
        else:
            for mol in supplier:
                mol_list.append(mol)

        self.sdf = mol_list

        return mol_list

    def feature_types(self, features: Optional[Union[str, dict]] = None):
        """
        Output feature types that will be displayed in the pharmacophore model. This will default to 'default' settings.
        However, default RDKit features can also be used or a dictionary containing a dictionary of features with SMARTS
        string can be used.
        :param features: Optional[Union[str, dict]]
            The feature type to be used in the pharmacophore model. This will default to "default" settings. Using
            "rdkit" will set the model to default RDKit features. Custom features can be used by giving a dictionary
            with feature:list pair, where the list contains SMARTS string for a given feature.

        :return:
        """
        # set instance variable
        if features is None:
            features = self.features

        global phrase
        if features == "default":
            phrase = f"Default features: \n('Donor', 'Acceptor', 'Aromatic', 'Hydrophobe')"
        elif features == "rdkit":
            phrase = f"Default features from RDKit: \n{feature_factory.GetFeatureFamilies()}"
        elif isinstance(features, dict):
            phrase = f"Custom features: \n{list(features.keys())}"

        return phrase

    def to_df(self, mols: list = None, mol_name: list = None, features: Optional[Union[str, dict]] = None):
        """
        From a list of containing Chem.MOl and a list containing molecule names, create a dataframe displaying matching
        features the molecules. Method will default to "default" settings. Users can use "rdkit" to set the model to
        default RDKit features, or give a dictionary with feature:list pair, where the list contains SMARTS string for a
        given feature.
        :param mols: list
            A list containing molecules in Chem.Mol format.
        :param mol_name: list
            A list containing names of molecules.
        :param features: Union[str, dict]
            The feature type to be used in the pharmacophore model. This will default to "default" settings. Using
            "rdkit" will set the model to default RDKit features. Custom features can be used by giving a dictionary
            with feature:list pair, where the list contains SMARTS string for a given feature.
        :return:
        """
        global feat_factory
        # set instance variable
        if features is None:
            features = self.features

        molecule_feature_frequencies = []

        # use default features
        if features == 'rdkit':
            feat_factory = feature_factory
            # get default features, count frequency, and append into list
            for mol in mols:
                feats = [feature.GetFamily() for feature in feat_factory.GetFeaturesForMol(mol)]
                feature_frequency = collections.Counter(feats)
                molecule_feature_frequencies.append(feature_frequency)

        # use default features
        elif features == 'default':
            feature_list = []
            # get features from constant file, count pharmacophore type, append to list
            for mol in mols:
                feats = self._calc_pharmacophore(mol)
                for feat in feats:
                    feature_list.append(feat[0])
                feature_frequency = collections.Counter(feature_list)
                molecule_feature_frequencies.append(feature_frequency)
                feature_list = []  # reset list to avoid double counting

        # use custom features
        elif isinstance(features, dict):
            # reset instance variable
            self.features = features

            feature_list = []
            # get features from constant file, count pharmacophore type, append to list
            for mol in mols:
                feats = self._calc_pharmacophore(mol, features)
                for feat in feats:
                    feature_list.append(feat[0])
                feature_frequency = collections.Counter(feature_list)
                molecule_feature_frequencies.append(feature_frequency)
                feature_list = []  # reset list to avoid double counting

        else:
            raise ValueError("Only 'default', 'rdkit', or custom features as dictionary are supported!")

        # # for troubleshooting
        # print(molecule_feature_frequencies)

        feature_frequencies_df = (
            pd.DataFrame(
                molecule_feature_frequencies,
                index=[f"Mol{i}" for i, _ in enumerate(mols, 1)]
            )
            .fillna(0)
            .astype(int)
        )

        # reformat table and rename columns to molecule list
        feature_frequencies_df = feature_frequencies_df.transpose()
        feature_frequencies_df = feature_frequencies_df.set_axis(mol_name, axis=1)

        return feature_frequencies_df

    def calc_pharm(self, mol: Chem.Mol = None, features: Optional[Union[str, dict]] = None):
        """
        Generate a list of pharmacophore features and position from a molecule.
        :param mol: Chem.Mol
            A molecule in ROMol format.
        :param features: str
            Designate the type of pharmacophore features to calculate from. Will default to "default" settings. This
            param can also accept "rdkit" to use RDKit default features or give a dictionary containing feature:list
            pair, where the list contains SMARTS string for a given feature.
        :return: list
            A list of pharmacophore features, matching atom index, and position from a molecule.
        """
        global pharmacophore
        # set instance variable
        if features is None:
            features = self.features

        # calculate pharmacophore by rdkit or pharmacophore dict
        if features == 'rdkit':
            pharmacophore = self._calc_rdkit(mol)
        elif features == 'default':
            pharmacophore = self._calc_pharmacophore(mol)
        elif isinstance(features, dict):
            pharmacophore = self._calc_pharmacophore(mol, features)
        else:
            raise ValueError("Only 'default', 'rdkit', or custom features as dictionary are supported!")

        # set features to self
        self.features_list = pharmacophore

        return pharmacophore

    def add_feats(self, mol: Chem.Mol = None, substruct: str = None, type: str = None):
        """
        Optional method to add specific features if not present using default or RDKit rules.
        :param mol: Chem.Mol
            Main molecule in ROMol format.
        :param substruct: str
            A SMARTS string for specific substructure to match
        :param type: str
            Set the type of interaction for substructure. Only 'Donor', 'Acceptor', 'Aromatic', and 'Hydrophobe' is
            allowed!
        :return:
        """
        # convert substruct smarts string into ROMol
        query = [Chem.MolFromSmarts(substruct)]

        # find matches
        matches = find_matches(mol, query)

        # check type
        if type == 'Donor' or type == 'Acceptor' or type == 'Aromatic' or type == 'Hydrophobe':
            pass
        else:
            raise ValueError(f"Type {type} not supported! Only 'Donor', 'Acceptor', 'Aromatic', and 'Hydrophobe' is "
                             f"allowed!")

        # obtain specific data from each match
        for match in matches:
            result = [type, match[0], match[1][0], match[1][1], match[1][2]]
            # add custom feats to self.features_list
            self.features_list.append(result)

        return self.features_list

    def output_features(self, feature_list: Optional[list] = None, savepath: str = None, sphere_size: float = 0.5,
                        color: dict = None):
        """
        Output features as a .pml format for visualization in PyMol.
        :param feature_list: Optional[list]
            A list containing features, corresponding atom number, and 3D position. Preferably genreated using the
            calc_pharm method.
        :param savepath: str = None
            Must be a file in .pml format.
        :param sphere_size: float = 0.5
            Set size of spheres.
        :param color: dict
            Set color for the pharmacophores. Must be given as Acceptor: color where color is a tuple for RGB or a
            string for a specific color to be translated into RGB format.
        :return:
        """
        if feature_list is None:
            feature_list = self.features_list

        with open(savepath, "w") as f:
            # feat_factory = feature_factory
            # features = feat_factory.GetFeaturesForMol(mol)
            print(f"Number of features: {len(feature_list)}")

            # set feature colors
            if color is None:
                for feat in FEATURE_COLORS:
                    f.write(f"set_color {feat}_color, {FEATURE_COLORS[feat]}\n")
            # if custom color, convert to RGB
            elif isinstance(color, dict):
                for shade in color:
                    rgb = color_convert(color[shade])
                    # print(rgb)  # for troubleshooting
                    f.write(f"set_color {shade}_color, {rgb}\n")
            else:
                warnings.warn("Issue with color features!")

            # to give sequential numbering for each group:
            feature_counts = {
                "Acceptor": 0, "Donor": 0, "Hydrophobe": 0, "Aromatic": 0, "LumpedHydrophobe": 0, "PosIonizable":0}

            # get features
            for feat in feature_list:
                feature = feat[0]  # extract feature type
                feature_counts[feature] += 1  # give feature count
                count = feature_counts[feature]  # get current count
                # get feature position
                pos_x = feat[2]
                pos_y = feat[3]
                pos_z = feat[4]

                f.write(
                    # f"pseudoatom {feature}_{count}, pos=[{pos_x}, {pos_y}, {pos_z}], color={color_type}\n"
                    f"pseudoatom {feature}_{count}, pos=[{pos_x}, {pos_y}, {pos_z}]\n"
                )

            # set color and sphere size
            f.write("show spheres, Acceptor_*\n")
            f.write("color acceptor_color, Acceptor_*\n")
            f.write("show spheres, Donor_*\n")
            f.write("color donor_color, Donor_*\n")
            f.write("show spheres, Hydrophobe_*\n")
            f.write("color hydrophobe_color, Hydrophobe_*\n")
            f.write("show spheres, Aromatic_*\n")
            f.write("color aromatic_color, Aromatic_*\n")
            f.write("show spheres, LumpedHydrophobe_*\n")
            f.write("color lumpedhydrophobe, LumpedHydrophobe_*\n")
            f.write("show spheres, PosIonizable*\n")
            f.write("color posionizable, PosIonizable*\n")
            f.write(f"set sphere_scale, {sphere_size}\n")  # Adjust sphere size in PyMOL

        print(f"Feature visualization script written to {savepath}.")

    def _calc_pharmacophore(self, mol: Chem.Mol = None, features: Optional[dict] = None):
        """
        Calculate pharmacophore features from a molecule using dict from constants
        :param mol: Chem.Mol
            Input molecule in ROMol format.
        :param features: Optional[dict]
            Include custom features for calculating features.
        :return:
            List of pharmacophore type and centroid coordinates.
        """
        global pharmacophore
        # read in custom feature dict
        if features is None:
            constant_feats = FEATURES
        else:
            constant_feats = features

        # hold matches
        matches = {}

        # Identify matches to feature dict
        for key, value in constant_feats.items():
            try:
                query = [Chem.MolFromSmarts(smarts) for smarts in value]
                matches[key] = find_matches(mol, query, verbose=False)
            except:
                pass
        # remove duplicate SMARTS matches
        cleaned_matches = {}
        for key, value in matches.items():
            unique_lists = []
            for list in value:
                if list not in unique_lists:
                    unique_lists.append(list)
            cleaned_matches[key] = unique_lists
        # create list containing pharmacophore and centroid coordinates
        pharmacophore = []
        for key, value in cleaned_matches.items():
            for match in value:
                # extarct feature type, atom match, and XYZ position
                p = [key, match[0], match[1][0], match[1][1], match[1][2]]
                pharmacophore.append(p)

        # remove duplicates for aromatic/hydrophobe set
        atom_indices = set()
        final_pharmacophore = []

        # extract aromatic/hydrophobe matches by atom index adn remove hydrophobe
        for entry in pharmacophore:
            label, atom_indexes, *rest = entry
            index = frozenset(atom_indexes)  # Convert atom indexes to a frozenset so order does not matter

            # Check if the atom indexes are already in the set
            if index not in atom_indices:
                final_pharmacophore.append(entry)
                atom_indices.add(index)
            # keep Aromatic
            elif label == 'Aromatic':
                final_pharmacophore = [e for e in final_pharmacophore if
                                       not (e[0] == 'Hydrophobe' and frozenset(e[1]) == index)]
                final_pharmacophore.append(entry)
                atom_indices.add(index)

        return final_pharmacophore

    def _calc_rdkit(self, mol: Chem.Mol = None):
        """
        Calculate pharmacophore features from a molecule using default RDKit methods.
        :param mol: Chem.Mol
            input molecule in ROMol format.
        :return:
            List of pharmacophore type and centroid coordinates.
        """
        global pharmacophore

        # load default RDKit feature factory file
        feat_factory = feature_factory

        # get list of features for molecule
        features = feat_factory.GetFeaturesForMol(mol)

        # store pharmacophore features as list
        pharmacophore = []
        for feature in features:
            fam = feature.GetFamily()
            pos = feature.GetPos()
            atom_indices = feature.GetAtomIds()
            pharmacophore_item = [fam, atom_indices, pos[0], pos[1], pos[2]]
            pharmacophore.append(pharmacophore_item)

        return pharmacophore


# Support function to fix bond order of molecule
def fix_bond_order(mol: Chem.Mol, template_smi: str, savepath: str = None):
    """
    Fix bond order for a given molecule. A template smiles string must be given.
    :param mol: Chem.Mol
        Molecule to fix. Must be in ROMol format.
    :param template_smi: str
        A smiles string for molecule to fix. Recommend canonical smiles string if possible.
    :param savepath: str
        Set the save path for the molecule. Output will be in sdf format.
    :return:
    """
    # remove hydrogen if present
    mol_prep = Chem.RemoveHs(mol)

    # read template
    template = Chem.MolFromSmiles(template_smi)

    # assign bond orders to crystal ligand
    fixed_mol = AllChem.AssignBondOrdersFromTemplate(mol_prep, template)

    # Map atoms from the template to the docked pose
    template_mapping = mol_prep.GetSubstructMatch(template)

    if not template_mapping:
        raise ValueError("Failed to match template to docked pose.")

    # Transfer 3D coordinates from docked_pose to fixed_mol
    new_conformer = Chem.Conformer(fixed_mol.GetNumAtoms())
    original_conformer = mol_prep.GetConformer()

    # set coordinates of heavy atoms to fixed_mol
    for new_idx, origin_idx in enumerate(template_mapping):
        pos = original_conformer.GetAtomPosition(origin_idx)
        new_conformer.SetAtomPosition(new_idx, pos)

    # Add 3D coordinates to the molecule
    fixed_mol.AddConformer(new_conformer, assignId=True)

    # save file
    Chem.MolToMolFile(fixed_mol, savepath)


def find_matches(mol: Chem.Mol = None, patterns: list[Chem.Mol] = None, verbose=True):
    """
    Support function to visualize matches between query molecule and features.
    :param mol: Chem.Mol
        Query molecule. Must be in ROMol format.
    :param patterns: list[Chem.Mol]
        A list of molecules to match against the query molecule.
    :param verbose: Bool
        Output messages for matches.
    :return:
    """
    matches = []
    for pattern in patterns:
        matched = mol.GetSubstructMatches(pattern)
        # get centroid and coordinates for each match
        for m in matched:
            try:
                centroid = _compute_match_centroid(mol, m)
                matches.append([m, centroid])
            except:
                pass
        # output statement if no matches
        if verbose is True:
            if len(matches) == 0:
                output_message = Chem.MolToSmarts(pattern)
                print(f"No Matches to {output_message}!")
                return matches

    return matches


def _compute_match_centroid(mol, matched_pattern):
    """
    Support function to calculate centroid of matches between query molecule and features.
    :param mol:
    :param matched_pattern:
    :return:
    """
    conf = mol.GetConformer()
    positions = [conf.GetAtomPosition(i) for i in matched_pattern]
    center = np.mean(positions, axis=0)
    # convert result to float
    center = center.tolist()
    return tuple(center)


if __name__ == "__main__":
    import doctest

    doctest.testmod()



"""
Script to draw pharmacophore in 2D figure
"""

import os
import io
import matplotlib.image as img
import matplotlib.pyplot as plt
from typing import Optional, Union
from PIL import Image
from collections import defaultdict
from cairosvg import svg2png
from IPython.display import SVG
from pharmacophore.constants import FEATURE_COLORS, color_convert
from pharmacophore import Pharmacophore
from rdkit import Chem
from rdkit.Chem import rdDepictor, AllChem
from rdkit.Chem.Draw import rdMolDraw2D, SimilarityMaps
from rdkit.Chem.Draw.MolDrawing import DrawingOptions


class Draw:
    def __init__(self, mol: Optional[Chem.Mol] = None, features: Optional[Union[str, dict]] = None):
        self.mol = mol
        self.features = features

    def draw_pharm(self, mol: Chem.Mol, features: Optional[Union[str, dict]] = None, color: dict = None,
                   savepath: str = None):
        """
        Draw a 2D pharmacophore image
        :param mol: Chem.Mol
            A molecule in ROMol format.
        :param features: str
            Determines which feature algorithm to use. Can use 'default' or 'rdkit'.
        :param color: dict
            Options to modify color of features. Must contain the name of the features. Colors can be given as either
            distinct name or hex codes.
        :param savepath: str
            Set the savepath to save the image.
        :return:
        """
        # set instance variables
        global color_palette
        if mol is None:
            mol = self.mol
        if features is None:
            features = self.features

        # calculate pharmacophore features
        try:
            pharm = Pharmacophore(features=features)
            calc_features = pharm.calc_pharm(mol=mol)
        except:
            raise ValueError('Unknown features! Only "default" or "rdkit", or custom features as dict are supported!')

        # dictionaries to highlight atoms and highlight radius
        atom_highlights = defaultdict(list)
        highlight_rads = {}

        # set to default or custom colors
        if color is None:
            color_palette = FEATURE_COLORS
        elif isinstance(color, dict):
            for key in color:
                color[key] = color_convert(color[key])
            color_palette = color

        # calc features and append results to atom_highlights and highlight_rads
        for feature in calc_features:
            if feature[0] in color_palette:
                # extract colors from constants
                color = color_palette[feature[0]]
                # # for troubleshooting
                # print(color)
                for atom_id in feature[1]:
                    atom_highlights[atom_id].append(color)
                    highlight_rads[atom_id] = 0.5

        # flatten molecule into 2D
        rdDepictor.Compute2DCoords(mol)
        rdDepictor.SetPreferCoordGen(True)
        drawer = rdMolDraw2D.MolDraw2DSVG(800, 800)

        # set drawing options
        # Use black for all elements
        drawer.drawOptions().updateAtomPalette(
            {k: (0, 0, 0) for k in DrawingOptions.elemDict.keys()}
        )
        drawer.SetLineWidth(2)
        drawer.SetFontSize(6.0)
        drawer.drawOptions().continuousHighlight = False
        drawer.drawOptions().splitBonds = False
        drawer.drawOptions().fillHighlights = True

        # get atom label
        for atom in mol.GetAtoms():
            atom.SetProp("atomLabel", atom.GetSymbol())
        drawer.DrawMoleculeWithHighlights(
            mol, "", dict(atom_highlights), {}, highlight_rads, {}
        )
        drawer.FinishDrawing()

        # draw molecule and save a temporary file
        svg = drawer.GetDrawingText().replace("svg:", "")
        SVG(svg)
        with open(f"pharm.svg", "w") as f:
            f.write(svg)

        # convert svg into png
        svg2png(bytestring=svg, write_to=f"image.png")

        # Set figure legend for feature and color type
        fig, (ax, picture) = plt.subplots(
            nrows=2,
            figsize=(4, 4),
            gridspec_kw={"height_ratios": [1, 5]},
        )

        # draw image and remove temporary file
        mol_image = img.imread(f"image.png")
        picture.imshow(mol_image)
        picture.axis("off")
        os.remove(f"image.png")
        os.remove(f"pharm.svg")

        # Data for the circles
        circle_radii = [0, 50, 100, 150, 200, 250]
        feature_values = list(color_palette.values())  # extract color values
        circle_colors = [i for i in feature_values]  # match circle to color values

        # get color labels for fig legend
        if isinstance(features, dict):
            circle_annotations = list(features.keys())
        elif features == "rdkit":
            circle_annotations = [
                "Donor",
                "Acceptor",
                "Aromatic",
                "Hydrophobe",
                "LumpedHydrophobe",
                "PosIonizable",
            ]
        else:
            circle_annotations = [
                "Donor",
                "Acceptor",
                "Aromatic",
                "Hydrophobe",
            ]
        # Draw the circles and annotations
        fontsize = 6
        if features == "rdkit":
            fontsize = 4

        for radius, color, annotation in zip(
                circle_radii, circle_colors, circle_annotations
        ):
            x = radius
            circle = plt.Circle((x, -5), 5, color=color)  # , alpha=0.5)
            ax.add_patch(circle)
            ax.annotate(
                annotation,
                (x, 10),
                va="center",
                ha="center",
                fontsize=fontsize,
                fontweight="bold",
            )

        # Set axis limits
        ax.set_xlim(-10, 270)
        ax.set_ylim(-20, 20)
        ax.axis("off")

        # Set aspect ratio to equal
        ax.set_aspect("equal", adjustable="box")
        if savepath:
            plt.savefig(f"{savepath}", dpi=300)

    # support function to draw molecule with atom index
    def atom_number(self, mol: Chem.Mol = None, label: str = "atomNote", size: tuple = (300, 300)):
        """
        Draw query molecule with labeled RDKit atom indices.
        :param mol: Chem.Mol
            A molecule in ROMol format.
        :param label: str
            Determines which style to label the molecule. Defaults to atomNote. In total, can use 'atomNote',
            'atomLabel', and 'molAtomMapNumber'.
        :param size: tuple
            Determine the size of the molecule to draw.
        :return:
        """
        # set instance variable
        if mol is None:
            mol = self.mol

        # Check if 2D coordinates are missing, and compute them if necessary. This will also strip away 3D coordinates.
        if not mol.GetNumConformers() or mol.GetConformer().Is3D():
            AllChem.Compute2DCoords(mol)

        # get atom indices
        for atom in mol.GetAtoms():
            if label == 'atomNote' or label == 'atomLabel' or label == 'molAtomMapNumber':
                atom.SetProp(label, str(atom.GetIdx()))
            else:
                raise ValueError("Only 'atomNote', 'atomLabel', and 'molAtomMapNumber' accepted!")

        # draw molecule
        img = Chem.Draw.MolToImage(mol, size=size)

        return img

    def similarity_maps(self, refmol: Chem.Mol = None, querymol: Chem.Mol = None, radius: int = 2, nbits: int = 2048,
                        fpType: str = 'bv', savepath: str = None):
        """
        Generate similarity map between query molecule and reference molecule. Only Morgan fingerprints are used.
        Image must be in png format.
        :param refmol: Chem.Mol
            Reference molecule. Must be in ROMol format.
        :param querymol: Chem.Mol
            Query molecule. Must be in ROMol format.
        :param radius: int
            Set the radius for the MorganFingerprint. Default to 2.
        :param nbits: int
            Set the number of bits for the generated fingerprint. Default to 2048.
        :param fpType: str
            Set the fingerprint type. Default to 'bv'. Can only use 'bv' or 'count'
        :param savepath: str
            Set image savepath.
        :return:
        """
        # set instance variable
        if refmol is None:
            refmol = self.mol

        # check fpType
        if fpType not in ['bv', 'count']:
            raise ValueError("Only 'bv' or 'count' accepted!")

        d = Chem.Draw.MolDraw2DCairo(400, 400)
        function = lambda m, i: SimilarityMaps.GetMorganFingerprint(m, i, radius=radius, fpType=fpType, nBits=nbits)
        _, maxWeight = SimilarityMaps.GetSimilarityMapForFingerprint(refMol=refmol, probeMol=querymol,
                                                                     fpFunction=function, draw2d=d)
        # finish drawing
        d.FinishDrawing()
        img_data = d.GetDrawingText()

        # draw image as png file
        bio = io.BytesIO(img_data)
        img = Image.open(bio)

        # if savepath given in params
        if savepath:
            img.save(savepath)

        return img


if __name__ == "__main__":
    import doctest

    doctest.testmod()
